// Name: Jonathan Turner
// Class: CS 4306/01
// Term: Spring 2024
// Instructor: Dr. Haddad
// Assignment: 2
// IDE Name: IntelliJ
/*
Algorithm Design Block
    Algorithm title: Finding Equal Disjoint Sets

    Logical steps:
        Step 1: Get the number of items in the set as user input.
        Step 2: Get the elements in the set as user input.
        Step 3: Add all the elements to the list and save it to a variable.
        Step 4: If the sum is odd, there is no disjoint sets and it is not possible.
        Step 5: Set a variable that will hold the first set and second set representing the disjoint and initialize it to null.
        Step 6: Create a List that holds a list of elements which represent each subset.
        Step 7: Generate the subsets using a subset generation method. (Ex. Bit Shift Counting)
        Step 8: Iterate through all the subsets that were generate.
        Step 9: If the current iteration's sum is equal to half of the total sum, move to step 11. Otherwise, continue iterating.
        Step 10: If the value is not found, then there is no disjoint equal subsets. End algorithm.
        Step 11: Assign the first set's variable to the current iteration's values.
        Step 12: Assign the second set's variable to the values of the original set that are not in the first set.
        Step 13: Return the results of the two equal disjoint sets that were found.

    Pseudocode Syntax:
        count <- input
        elements <- input
        first_set, second_set <- empty

        sum <- 0
        for each number in elements:
            sum <- sum + number

        if sum % 2 is 0:
            subsets <- generate_subsets(elements)
            for each subset in subsets:
                current_sum <- 0
                for each value in subset:
                    current_sum <- current_sum + value
                if sum / 2 is current_sum:
                    first_set <- subset

            if first_set is not empty:
                second_set <- elements - first_set
            else
                return "no equal disjoint subsets"

        return first_set, second_set

    Big-O Analysis: (Based on Implementation)

    Best-Case Scenario:
    Worse-Case Scenario:
    Average-Case:

    Big O =>
*/
package Assignments.A2;

import java.util.ArrayList;
import java.util.List;

/**
 * Generates, if possible, two equal subsets of a given set of elements that are equal to each other.
 *
 * @author Jonathan Turner
 * @version Spring 2024
 */
public class Partition {

    private int[] set;
    private List<List<Integer>> subsets;

    /**
     * Iterates through all the possible subsets generated by this.generateSubsets
     * and if a subset is found that is exactly half the size it sets the firstset
     * to those values and the second subset to the missing values.
     *
     * @precondition set.length != 0
     * @postcondition the resulting sets are printed.
     */
    public void performAlgorithm() {
        int totalSum = 0;
        for (int i = 0; i < this.set.length; i++) {
            totalSum += this.set[i];
        }
        List<Integer> firstSubset = null;
        List<Integer> secondSubset = null;

        /*
         * Checks through every subset that was generated,
         * if a subset is found it sets the firstSubset to the found subset.
        */
        for (List<Integer> subset : subsets) {
            int subsetTotal = 0;
            for (int value : subset) {
                subsetTotal += value;
            }
            if (totalSum / 2 == subsetTotal) {
                firstSubset = subset;
                break;
            }
        }

        /* Checks if a list was found, if so creates second sublist. */
        if (firstSubset != null) {
            secondSubset = new ArrayList<>();
            for (int secondValue : this.set) {
                if (!firstSubset.contains(secondValue)) {
                    secondSubset.add(secondValue);
                }
            }
        }
        printResults(firstSubset, secondSubset);
    }

    /**
     * Prints the results of the algorithm based on the two subsets. Prints the setsize, the original set.
     * If the subsets were null (No match found) print output associated, else print the both lists.
     *
     * @precondition none
     * @postcondition the output is displayed.
     *
     * @param firstSubset the first subset.
     * @param secondSubset the second subset.
     */
    private void printResults(List<Integer> firstSubset, List<Integer> secondSubset) {
        int setSize = this.set.length;
        String original = "";
        for (int values : this.set) {
            original += values + " ";
        }
        String first = "{";
        String second = "{";
        if (firstSubset != null) {
            for (int fvalues : firstSubset) {
                first += fvalues + ",";
            }
            first = first.substring(0, first.length()-1) + "}";
            for (int svalues : secondSubset) {
                second += svalues + ",";
            }
            second = second.substring(0, second.length()-1) + "}";
        } else {
            first = "No disjoint subsets with the same sum";
            second = "of their elements found";
        }

        System.out.println("Set size:               " + setSize);
        System.out.println("Integer values:         " + original);
        System.out.println("Disjoint subsets with same sum: " + first);
        System.out.println("                                " + second);
    }

    /**
     * Generates the list of subsets using Bit Shift Counting by iterating over all combinations of 1 to n.
     * After shifting the counter, the current subset is created initialized and if 1 shifted by j is logically
     * ANDed with the i is not 0, it adds the value j represents to the subset.
     *
     * @precondition none
     * @postcondition the subsets are generated and added to the field.
     */
    public void generateSubsets() {
        List<List<Integer>> subsets = new ArrayList<>();
        int n = this.set.length;

        for (int i = 0; i < (1 << n); i++) {
            List<Integer> currentSubset = new ArrayList<>();
            for (int j = 0; j < n; j++) {
                if ((i & (1 << j)) != 0) {
                    currentSubset.add(this.set[j]);
                }
            }
            subsets.add(currentSubset);
        }
        this.subsets = subsets;
    }

}
